#!/usr/bin/env python3
import os
import glob
import sys
import datetime
import re

def getTimestamp(yyyy, MMdd, hhmmss):
    #print(yyyy, MMdd, hhmmss)

    year   = int(yyyy)
    if year < 2000:
        print("Error!  Invalid year, quitting program.")
        sys.exit(1)

    month  = int(MMdd[:2])
    day    = int(MMdd[3:])
    hour   = int(hhmmss[:2])
    minute = int(hhmmss[3:5])
    second = int(hhmmss[6:])

    #print(year, month, day, hour, minute, second)

    return datetime.datetime(year, month, day, hour, minute, second).timestamp()

def getDateString(yyyy, MMdd, hhmmss):
    #print(yyyy, MMdd, hhmmss)
    return MMdd + "/" + yyyy + " " + hhmmss

class License:
    def __init__(self, uniqueID, checkoutTime, checkoutString, product):
        self.uniqueID = uniqueID
        self.checkoutTime = checkoutTime
        self.checkoutString = checkoutString
        self.product = product
        self.checkinTime = None
        self.checkinString = None
        self.checkoutDuration = None

filesToParse = sys.argv[1:]

productList = ['barracuda-15', 'barracuda-15-chem', 'barracuda-gpu']

outList = []

outCount = {}
outCountFiles = {}

for p in productList:
    outCount[p] = 0
    outCountFiles[p] = open("license_count_vs_time_" + p + ".csv", 'w')
    outCountFiles[p].write("Date,License Count\n")

#print(outCount)

# Keep track of calendar days when more than 1 solver license was in use
daysWithMultipleLicensesInUse = set()

# In some cases, licenses are never checked back in.  Set the parameters below
# to enable this script to consider a license as forcefully checked in after
# "forceAfterN" days.
forceCheckin = True
forceAfterN = 60  # days

csvFile = "license_usage_report.csv"
csv = open(csvFile, 'w')
csv.write("Product,User,Host,Handle,Checkout,Checkin,Duration (s)\n")

for fileName in filesToParse:
    myYear = -1

    with open(fileName, mode='r', encoding="utf8", errors='ignore') as f:
        lineNumber = 1
        for line in f:
            splitLine = line.split()
            if len(splitLine) > 0:
                # Get the year by looking at the "START" line
                # START ls58.rlmcloud.com 06/01/2022 00:00:29
                if splitLine[0] == "START":
                    myDate = splitLine[2]
                    myYear = myDate[6:]
                    #print(lineNumber, myYear)

                datePattern = re.compile('\d\d/\d\d/\d\d\d\d')
                if re.match(datePattern, splitLine[0]):
                    myDate = splitLine[0]
                    myYear = myDate[6:]
                    #print(lineNumber, myYear)

                for myProduct in productList:
                    # Look for "OUT" lines like this:
                    # OUT barracuda-15 2022.04 1 user host "" 1 1 0 489 489 1df82 "" "" "" 06/07 09:03:59
                    #
                    # Based on the RLM License Administrator Manual section "Reportlog File Format":
                    # (splitLine index)  0   1       2       3     4    5    6         7     8       9          10            11           12         13        14                  15                  16    17
                    # "standard" format: OUT product version pool# user host "isv_def" count cur_use cur_resuse server_handle share_handle process_id "project" "requested product" "requested version" mm/dd hh:mm:ss
                    if splitLine[0] == "OUT" and splitLine[1] == myProduct:
                        #print(line)
                        product = splitLine[1]
                        version = splitLine[2]
                        user    = splitLine[4]
                        host    = splitLine[5]
                        count   = splitLine[7]
                        handle  = splitLine[10]
                        MMdd    = splitLine[16]
                        hhmmss  = splitLine[17]
                        #print("Parsed OUT", product, version, user, host, count, handle, MMdd, hhmmss)

                        timestamp = getTimestamp(myYear, MMdd, hhmmss)
                        #print(timestamp)

                        uniqueID = '+'.join([product, version, user, host, handle])
                        #print(uniqueID)

                        for item in outList:
                            #print("Item uniqueID:", item.uniqueID)
                            if uniqueID == item.uniqueID:
                                #print("Warning: license checked out multiple times", uniqueID)
                                continue
                        
                        outList.append(License(uniqueID, timestamp, getDateString(myYear, MMdd, hhmmss),product))
                        outCountFiles[product].write(getDateString(myYear, MMdd, hhmmss) + "," + str(outCount[product]) + "\n")
                        outCount[product] += 1
                        #print(outCount)
                        outCountFiles[product].write(getDateString(myYear, MMdd, hhmmss) + "," + str(outCount[product]) + "\n")

                        if product == "barracuda-15" and outCount[product] > 1:
                            daysWithMultipleLicensesInUse.add(myYear + "/" + MMdd)
 
                    # Look for "IN" lines like this:
                    # IN 1 barracuda-15 2021.11 user host "" 1 0 0 159 05/05 08:25:14
                    #
                    # Based on the RLM License Administrator Manual section "Reportlog File Format":
                    # (splitLine index)  0  1   2       3       4    5    6         7     8       9          10            11    12
                    # "standard" format: IN why product version user host "isv_def" count cur_use cur_resuse server_handle mm/dd hh:mm:ss
                    if splitLine[0] == "IN" and splitLine[2] == myProduct:
                        #print(line)
                        product = splitLine[2]
                        version = splitLine[3]
                        user    = splitLine[4]
                        host    = splitLine[5]
                        count   = splitLine[7]
                        handle  = splitLine[10]
                        MMdd    = splitLine[11]
                        hhmmss  = splitLine[12]
                        #print("Parsed IN", product, version, user, host, count, handle, MMdd, hhmmss)

                        try:
                            timestamp = getTimestamp(myYear, MMdd, hhmmss)
                            #print(timestamp)
                        except:
                            print("Error when converting timestamp.")
                            print("Input file:", fileName)
                            print("Line number:", lineNumber)
                            print("Line:", line)
                            print("myYear, MMdd, hhmmss:", myYear, MMdd, hhmmss)
                            sys.exit(-1)

                        uniqueID = '+'.join([product, version, user, host, handle])
                        #print(uniqueID)

                        for item in outList:
                            #print("Item uniqueID:", item.uniqueID)

                            staleCheckout = False
                            if forceCheckin == True:
                                myCheckoutDuration = timestamp - item.checkoutTime

                                if myCheckoutDuration >= forceAfterN * 3600 * 24:
                                    #print("Forcefully checking in license", item.uniqueID, myCheckoutDuration)
                                    staleCheckout = True

                            if uniqueID == item.uniqueID or staleCheckout == True:
                                #print("Found uniqueID match!", uniqueID)

                                item.checkinTime = timestamp
                                item.checkinString = getDateString(myYear, MMdd, hhmmss)
                                item.checkoutDuration = item.checkinTime - item.checkoutTime

                                #print(uniqueID, item.checkinTime, "{0:.2f}".format(item.checkoutDuration/3600/24), "days")

                                outList.remove(item)
                                #print(outList)
                                #print("--------------------------------------------------------------")

                                outCountFiles[item.product].write(item.checkinString + "," + str(outCount[item.product]) + "\n")
                                outCount[item.product] -= 1
                                #print(outCount)
                                outCountFiles[item.product].write(item.checkinString + "," + str(outCount[item.product]) + "\n")

                                csv.write(product + "," + user + "," + host + "," + handle + "," + item.checkoutString + "," + item.checkinString + "," + "{0:.2f}".format(item.checkoutDuration) + "\n")

            lineNumber += 1

#print("Number of licenses not checked back in =", len(outList))
#
#for item in outList:
#    print(item.uniqueID, item.checkoutString)

for p in productList:
    outCountFiles[p].close()

csv.close()

#print(daysWithMultipleLicensesInUse)
print("Number of calendar days with >1 solver license in use =", len(daysWithMultipleLicensesInUse))

print("License usage analysis script finished running.  See the following files for results:")
print("license_usage_report.csv")
print("license_count_vs_time_barracuda-15.csv")
print("license_count_vs_time_barracuda-15-chem.csv")
print("license_count_vs_time_barracuda-15-gpu.csv")
